/**
  * <!-- LICENSE_TEXT_START -->
* Copyright 2007-2010 SAIC. This software was developed in conjunction with the National Cancer Institute,
* and so to the extent government employees are co-authors, any rights in such works shall be subject to Title 17 of the United States Code, section 105.
* Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the disclaimer of Article 3, below. Redistributions
* in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other
* materials provided with the distribution.
* 2. The end-user documentation included with the redistribution, if any, must include the following acknowledgment:
* "This product includes software developed by the SAIC and the National Cancer Institute."
* If no such end-user documentation is to be included, this acknowledgment shall appear in the software itself,
* wherever such third-party acknowledgments normally appear.
* 3. The names "The National Cancer Institute", "NCI" and "SAIC" must not be used to endorse or promote products derived from this software.
* 4. This license does not authorize the incorporation of this software into any third party proprietary programs. This license does not authorize
* the recipient to use any trademarks owned by either NCI or SAIC-Frederick.
* 5. THIS SOFTWARE IS PROVIDED "AS IS," AND ANY EXPRESSED OR IMPLIED WARRANTIES, (INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE) ARE DISCLAIMED. IN NO EVENT SHALL THE NATIONAL CANCER INSTITUTE,
* SAIC, OR THEIR AFFILIATES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * <!-- LICENSE_TEXT_END -->
  */
/**
  * @author EVSAPI Team
  * @version 4.0
 */
package gov.nih.nci.codegen.transformer;

import gov.nih.nci.codegen.ArtifactHandler;
import gov.nih.nci.codegen.GenerationException;
import gov.nih.nci.codegen.GeneratorError;
import gov.nih.nci.codegen.GeneratorErrors;
import gov.nih.nci.codegen.Transformer;
import gov.nih.nci.codegen.artifact.BaseArtifact;
import gov.nih.nci.codegen.util.Constant;
import gov.nih.nci.codegen.util.TransformerUtils;
import gov.nih.nci.ncicb.xmiinout.domain.UMLAssociationEnd;
import gov.nih.nci.ncicb.xmiinout.domain.UMLAttribute;
import gov.nih.nci.ncicb.xmiinout.domain.UMLClass;
import gov.nih.nci.ncicb.xmiinout.domain.UMLModel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 *
 * @author caCORE SDK Team
 * @version 1.0
 */
public class XmlMappingTransformer implements Transformer { 


	protected Logger log = Logger.getLogger(XmlMappingTransformer.class);
	
	protected GeneratorErrors generatorErrors = new GeneratorErrors();
	
	private ArtifactHandler artifactHandler;
	
	private String namespaceUriPrefix;

	private boolean includeAssociations = false;    
	private boolean includeFieldHandler = false;    

	private boolean enabled = true;
	
	private String name = XmlMappingTransformer.class.getName();	

	
	/* @param model The UMLModel containing the classes for which a 
	 * Castor Mapping file should be generated
	 * @see gov.nih.nci.codegen.Transformer#execute(gov.nih.nci.ncicb.xmiinout.domain.UMLModel)
	 */
	public GeneratorErrors execute(UMLModel model)
	{
		log.info("Executing " + getName());		
		
		log.debug("Model name: " + model.getName());

		GeneratorErrors errors = new GeneratorErrors();
		Collection<UMLClass> classes = null;

		try {
			classes = TransformerUtils.getAllClasses(model);
		} catch(GenerationException ge){
			errors.addError(new GeneratorError(getName() + ": Error while retrieving classes to process: ", ge));
		}
		
		log.debug("classes size: " + classes.size());		
		processClasses(classes);
		
		return generatorErrors;
	}

	/* @param model The UMLModel containing the classes for which a 
	 * Castor Mapping file should be generated.  This method will run any
	 * necessary validations.
	 * @see gov.nih.nci.codegen.Transformer#validate(gov.nih.nci.ncicb.xmiinout.domain.UMLModel)
	 */
	public GeneratorErrors validate(UMLModel model)
	{
		return null;
	}	

	/**
	 * @param classColl The collection of classes for which
	 * a Castor Mapping file will be generated
	 */
	private void processClasses(Collection<UMLClass> classColl)
	{

		Document doc = generateMappingDoc(classColl);
		XMLOutputter p = new XMLOutputter();
		p.setFormat(Format.getPrettyFormat());	

		BaseArtifact artifact = new BaseArtifact();
		artifact.setContent(p.outputString(doc));

		try {
			artifactHandler.handleArtifact(artifact);
		} catch(GenerationException ge) {
			log.error("ERROR: ", ge);
			
		}        
	}

	/**
	 * @param classColl The List of classes for which a Castor Mapping file should be generated.
	 * @return The JDOM Document for the Castor Mapping file.
	 */
	private Document generateMappingDoc(Collection<UMLClass> classColl) {

		Element mappingEl = new Element("mapping");
		DocType docType = new DocType("mapping","-//EXOLAB/Castor Object Mapping DTD Version 1.0//EN","http://www.castor.org/mapping.dtd");

		Document doc = new Document();
		doc.setDocType(docType);
		Collection sortedClassifier = sortClasses(classColl);
		for (Iterator i = sortedClassifier.iterator(); i.hasNext();) {
			UMLClass klass = (UMLClass) i.next();
			doMapping(klass, mappingEl);
		}
		doc.setRootElement(mappingEl);
		return doc;
	}

	/**
	 * @param classColl The List of classes to be sorted.  The caCOREMarshaller 
	 * has problem with forward references.
	 * @return The Sorted list of classes with Generalizations listed first.
	 */
	private Collection sortClasses(Collection<UMLClass> classColl) 
	{
		// The caCOREMarshaller has problem with forward references. Sort
		// the class with the list generalizations to the top.
		class caCOREComparator implements Comparator {
			public int compare(Object obj1, Object obj2)
			{
				int weight = determineWeight((UMLClass)obj1) - determineWeight((UMLClass)obj2);
				log.debug("weight: " + weight);
				return weight;
			}

			private int determineWeight(UMLClass obj)
			{
				int count = -1;
				UMLClass superClass = obj;
				try {
					do {
						superClass = TransformerUtils.getSuperClass(superClass);
						count++;
					} while (superClass!=null);

				} catch(GenerationException ge){
					log.error("ERROR: ", ge);
					generatorErrors.addError(new GeneratorError(getName() + ": " + ge.getMessage(), ge));
				}
				return count;
			}
		};

		ArrayList<UMLClass> list = new ArrayList<UMLClass>(classColl);
		Collections.sort(list, new caCOREComparator());
		return list;
	}

	/**
	 * @param klass The UML Class for which a Namespace URI will be generated
	 * @return The Namespace URI generated for the UML Class 
	 */
	private String getNamespaceURI(UMLClass klass) {
		String packageName = TransformerUtils.getFullPackageName(klass);
		
		StringBuffer nsURI = new StringBuffer();
		nsURI.append(namespaceUriPrefix);
		nsURI.append(packageName);
		return nsURI.toString();
	}

	/**
	 * @param klass The UMLClass for which a Mapping Element should be generated
	 * @param mappingEl The Mapping Element to which the Class information will be added.
	 */
	private void doMapping(UMLClass klass, Element mappingEl) { 
//		String superClassName =null;
		
		UMLClass superClass = null;
		try {
			superClass = TransformerUtils.getSuperClass(klass);
		} catch (GenerationException ge){
			log.error("ERROR: ", ge);
			generatorErrors.addError(new GeneratorError(getName() + ": " + ge.getMessage(), ge));
		}	

//		if (superClass != null) {
//			superClassName = TransformerUtils.getFullPackageName(superClass)+ Constant.DOT+superClass.getName();
//		} 	    
		String classElName = "class";
		Element classEl = new Element(classElName);
		mappingEl.addContent(classEl);
		classEl.setAttribute("name", TransformerUtils.getFullPackageName(klass)+Constant.DOT+klass.getName());
		
		/*
		 //Commented idAttribute: Shaziya
		String idAttrName = "";
		try {
			idAttrName = TransformerUtils.getClassIdAttr(klass).getName();
			log.debug("className: " + klass.getName() + "; idAttrName: " + idAttrName);
		} catch (GenerationException ge){
			log.error("ERROR: ", ge);
			generatorErrors.addError(new GeneratorError(getName() + ": " + ge.getMessage(), ge));
		}
				
		classEl.setAttribute("identity", idAttrName);
		if (superClassName!=null){
			classEl.setAttribute("extends", superClassName);
		}
		*/
		Element maptoelement = new Element("map-to");
		maptoelement.setAttribute("xml", klass.getName());
		String nsURI = getNamespaceURI(klass);
		maptoelement.setAttribute("ns-uri",nsURI);
		classEl.addContent(maptoelement);

		//Do properties
		for (Iterator i = klass.getAttributes().iterator(); i.hasNext();) {
			UMLAttribute att = (UMLAttribute) i.next();
			Element field = new Element("field");
			field.setAttribute("name", att.getName());
			log.debug("Field name: " + att.getName());

			String type = TransformerUtils.getType(att);
			String qName = getQualifiedTypeName(type);
			if (qName.equalsIgnoreCase("collection")) {
				log.debug("Handling type 'collection' - qName: " + qName);            	 
				field.setAttribute("type", "string");
				field.setAttribute("collection", qName);
				Namespace namespace = Namespace.getNamespace(qName,nsURI);

				Element bind = new Element("bind-xml");
				bind.setAttribute("name", qName + ":" + att.getName());
				bind.setAttribute("QName-prefix",qName,namespace);
				bind.setAttribute("node", "attribute");
				field.addContent(bind);
			} else {
				field.setAttribute("type", getQualifiedTypeName(TransformerUtils.getType(att)));
				Element bind = new Element("bind-xml");
				bind.setAttribute("name", att.getName());
				bind.setAttribute("node", "attribute");
				field.addContent(bind);         	 
			}
			
			classEl.addContent(field);
		}

		if (includeAssociations) {
			log.debug("*********** klass: " + klass.getName());
			log.debug("*********** TransformerUtils.getAssociationEnds(klass).size(): " + TransformerUtils.getAssociationEnds(klass).size());

			for (Iterator i = TransformerUtils.getAssociationEnds(klass).iterator(); i.hasNext();) {
				UMLAssociationEnd thisEnd = (UMLAssociationEnd) i.next();
				UMLAssociationEnd otherEnd = TransformerUtils.getOtherAssociationEnd(thisEnd);
				addSequenceAssociationElement(classEl, thisEnd, otherEnd);
			}
		}

	}

	/**
	 * @param mappingEl The Mapping Element to which content should be added
	 * @param thisEnd The primary end of the UMLAssociation
	 * @param otherEnd The other end of the UMLAssociation
	 */
	private void addSequenceAssociationElement(Element mappingEl, UMLAssociationEnd thisEnd, UMLAssociationEnd otherEnd) { 
		if (otherEnd.isNavigable()) {
			// If classes belong to the same package then do not qualify the association

			String otherEndTypeName = ((UMLClass)(otherEnd.getUMLElement())).getName();
			log.debug("mappingEl.getName(): " + mappingEl);
			log.debug("thisEnd.getRoleName(): " + thisEnd.getRoleName());
			log.debug("thisEnd.getType().getName(): " + ((UMLClass)(thisEnd.getUMLElement())).getName());    		 
			log.debug("otherEnd.getRoleName(): " + otherEnd.getRoleName()); 
			log.debug("otherEnd.getType().getName(): " + otherEndTypeName);  

			if (TransformerUtils.isMany2One(thisEnd, otherEnd)) {

				log.debug("UML13Utils.isMany2One(thisEnd, otherEnd): " + true);
				log.debug("lowerBound: " + TransformerUtils.getLowerBound(otherEnd));
				log.debug("upperBound: " + TransformerUtils.getUpperBound(otherEnd));

				Element field = new Element("field");
				field.setAttribute("name", otherEnd.getRoleName() ); //otherEnd.getName());
				String associationPackage = TransformerUtils.getFullPackageName((UMLClass)otherEnd.getUMLElement());
				field.setAttribute("type", associationPackage + Constant.DOT + ( (UMLClass)otherEnd.getUMLElement() ).getName() );
				if (includeFieldHandler) {
					field.setAttribute("handler", "gov.nih.nci.system.client.util.xml.CastorDomainObjectFieldHandler" );
				}
				Element bind = new Element("bind-xml");
				bind.setAttribute("name", otherEndTypeName );
				bind.setAttribute("type", associationPackage + Constant.DOT + otherEndTypeName );
				bind.setAttribute("location", otherEnd.getRoleName() ); //otherEnd.getName());	                 
				bind.setAttribute("node", "element");
				field.addContent(bind);

				mappingEl.addContent(field);

			} else if (TransformerUtils.isMany2Many(thisEnd, otherEnd) || TransformerUtils.isOne2Many(thisEnd, otherEnd)){
				log.debug("UML13Utils.isMany2Many(thisEnd, otherEnd): " + TransformerUtils.isMany2Many(thisEnd, otherEnd));
				log.debug("UML13Utils.isOne2Many(thisEnd, otherEnd): " + TransformerUtils.isOne2Many(thisEnd, otherEnd));
				log.debug("lowerBound: " + TransformerUtils.getLowerBound(otherEnd));
				log.debug("upperBound: " + TransformerUtils.getUpperBound(otherEnd));
				Element field = new Element("field");
				field.setAttribute("name", otherEnd.getRoleName() ); //otherEnd.getName());
				String associationPackage = TransformerUtils.getFullPackageName((UMLClass)otherEnd.getUMLElement());
				field.setAttribute("type", associationPackage + Constant.DOT + otherEndTypeName);
				field.setAttribute("collection", "collection" );
				if (includeFieldHandler) {
					field.setAttribute("handler", "gov.nih.nci.system.client.util.xml.CastorCollectionFieldHandler" );
				}

				//for container = false
				//field.setAttribute("container", "false" );

				Element bind = new Element("bind-xml");
				//bind.setAttribute("auto-naming", "deriveByClass");

				// for container = false
				//bind.setAttribute("name", otherEnd.getName());


				// for container = true
				bind.setAttribute("name", otherEndTypeName );
				bind.setAttribute("type", associationPackage+Constant.DOT + otherEndTypeName);
				bind.setAttribute("location", otherEnd.getRoleName() ); //otherEnd.getName());
				bind.setAttribute("node", "element");

				field.addContent(bind);
				mappingEl.addContent(field);

			}

		}
	}

	/**
	 * @param type The UMLAttribute type
	 * @return The corresponding Castor Mapping attribute type
	 */
    //Modified to support EVS related data types : Shaziya
	private String getQualifiedTypeName(String type) {

		if ("Character".equalsIgnoreCase(type) || "char".equalsIgnoreCase(type)) {
			return "char";
		}

		if ("HashSet".equalsIgnoreCase(type)) {
			return "java.util.Set";
		}

		if ("HashMap".equalsIgnoreCase(type)) {
			return "java.util.Map";
		}
        if ("Vector".equalsIgnoreCase(type)) {
            return "java.util.Vector";
        }
        if ("ArrayList".equalsIgnoreCase(type)) {
            return "java.util.ArrayList";
        }
        if ("Set".equalsIgnoreCase(type)) {
            return "java.util.Set";
        }
        if ("Map".equalsIgnoreCase(type)) {
            return "java.util.Map";
        }

		log.debug("*** type: " + type);
		return type.toLowerCase();
	}

	/**
	 * @param includeAssociations set the flag to indicate whether or not 
	 * associations should be processed during the Mapping generation
	 */
	public void setIncludeAssociations(boolean includeAssociations) {
		this.includeAssociations = includeAssociations;
	}

	/**
	 * @param includeFieldHandler set the flag to indicate whether or not a 
	 * Castor Mapping Field Handler should be generated
	 */
	public void setIncludeFieldHandler(boolean includeFieldHandler) {
		this.includeFieldHandler = includeFieldHandler;
	}

	/**
	 * @param artifactHandler the artifactHandler to set; called by the
	 *        Spring framework
	 */
	public void setArtifactHandler(ArtifactHandler artifactHandler)
	{
		this.artifactHandler = artifactHandler;
	}

	/**
	 * Sets the namespace uri prefix.
	 * 
	 * @param namespaceUriPrefix the new namespace uri prefix
	 */
	public void setNamespaceUriPrefix(String namespaceUriPrefix) {
		this.namespaceUriPrefix = namespaceUriPrefix;
	}


	/**
	 * Sets the enabled value of the schema.
	 * 
	 * @param enabled a boolean value
	 */
	public void setEnabled(boolean enabled)
	{
		this.enabled = enabled;
	}
	
	/* (non-Javadoc)
	 * @see gov.nih.nci.codegen.Transformer#isEnabled()
	 */
	public Boolean isEnabled() {
		return enabled;
	}

	/* (non-Javadoc)
	 * @see gov.nih.nci.codegen.Transformer#getName()
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets the name.
	 * 
	 * @param name the new name
	 */
	public void setName(String name) {
		this.name = name;
	}

}
