/*
 * Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 *
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * 		http://www.eclipse.org/legal/epl-v10.html
 *
 */

 /*
 * Modified by Kim Ong, NGIT for supporting NCI Bioportal
 */

package gov.nih.nci.lexbig.ext;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import org.LexGrid.LexBIG.DataModel.Collections.AssociatedConceptList;
import org.LexGrid.LexBIG.DataModel.Collections.AssociationList;
import org.LexGrid.LexBIG.DataModel.Collections.NameAndValueList;
import org.LexGrid.LexBIG.DataModel.Collections.ResolvedConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.AssociatedConcept;
import org.LexGrid.LexBIG.DataModel.Core.Association;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Extensions.Generic.LexBIGServiceConvenienceMethods;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.LexGrid.LexBIG.Utility.ConvenienceMethods;
import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.naming.SupportedAssociation;
import org.apache.log4j.Logger;

//[#12808] LexBIG 2.2 beta compatibility.
import org.LexGrid.LexBIG.Impl.Extensions.GenericExtensions.LexBIGServiceConvenienceMethodsImpl;


/**
 * Example showing how to determine and display an unsorted list of root and
 * subsumed nodes, up to a specified depth, for hierarchical relationships.
 * It is primarily geared towards the NCI Metathesaurus source.
 *
 * Hierarchies are navigated primarily through the UMLS-defined PAR (has parent)
 * and CHD (has child) relationships.
 *
 * This program accepts two parameters.  The first indicates the depth
 * to display hierarchical relations.  If 0, only the root nodes are displayed.
 * If 1, nodes immediately subsumed by the root are also displayed, etc.
 * If < 0, a default depth of 0 is assumed.
 *
 * The second parameter must provide the source abbreviation (SAB) of the
 * Metathesaurus source to be evaluated (e.g. ICD9CM, MDR, SNOMEDCT).
 */
public class NavigationUtils {
	final static int DEFAULT_DEPTH = 0;

	/** The Constant BROADENING_HIER_ASSOC. */
	public final static String[] BROADENING_HIER_ASSOC = new String[] {"CHD", "hasSubtype", "hasSubclass", "inverse_isa", "inverse_is_a"};

	/** The Constant NARROWING_HIER_ASSOC. */
	public final static String[] NARROWING_HIER_ASSOC = new String[] {"PAR", "RB", "isa", "is_a", "branch_of", "part_of", "tributary_of"};

//	private static Map<String, String> assocFwdNames_ = new HashMap<String, String>();
//	private static Map<String, String> assocRevNames_ = new HashMap<String, String>();

    private static PrintWriter pw = null;

   	private static Logger log = Logger.getLogger("NavigationUtils");

	/**
	 * Instantiates a new navigation utils.
	 */
	public NavigationUtils() {
		super();
	}

	/**
	 * Gets the root nodes.
	 *
	 * @param lbSvc the LexBIGService
	 * @param sab the sab
	 * @param scheme the scheme
	 * @param csvt the CodingSchemeVersionOrTag
	 *
	 * @return the root nodes as a ResolvedConceptReferenceList
	 *
	 * @throws LBException the LB exception
	 */
	public static ResolvedConceptReferenceList getRootNodes(
		    LexBIGService lbSvc,
			String sab,
			String scheme,
			CodingSchemeVersionOrTag csvt) throws LBException
	{
		HashSet<String> hset = new HashSet<String>();
        ResolvedConceptReferenceList list = new ResolvedConceptReferenceList();
		// Find the root nodes of the hierarchy (attached to '@' or '@@' nodes).
		String[] assocNames = BROADENING_HIER_ASSOC;
		for (int i = 0; i < assocNames.length; i++)
		{
			try {
				ResolvedConceptReferenceList roots = getTopNodes(lbSvc, scheme, csvt, null, assocNames[i], sab);
				if (roots != null && roots.getResolvedConceptReferenceCount() > 0)
				{
					for (Iterator<ResolvedConceptReference> nodes = roots.iterateResolvedConceptReference(); nodes.hasNext(); )
					{
						ResolvedConceptReference rcr = (ResolvedConceptReference) nodes.next();
						String code = rcr.getReferencedEntry().getConceptCode();
						if (!hset.contains(code))
						{
							hset.add(code);
							list.addResolvedConceptReference(rcr);
						}
					}

				}
			} catch (LBException lbe) {};
		}

		assocNames = NARROWING_HIER_ASSOC;
		for (int i = 0; i < assocNames.length; i++)
		{
			try {
				ResolvedConceptReferenceList roots = getEndNodes(lbSvc, scheme, csvt, null, assocNames[i], sab);
				if (roots != null && roots.getResolvedConceptReferenceCount() > 0)
				{
					for (Iterator<ResolvedConceptReference> nodes = roots.iterateResolvedConceptReference(); nodes.hasNext(); )
					{
						ResolvedConceptReference rcr = (ResolvedConceptReference) nodes.next();
						String code = rcr.getReferencedEntry().getConceptCode();
						if (!hset.contains(code))
						{
							hset.add(code);
							list.addResolvedConceptReference(rcr);
						}
					}

				}
			} catch (LBException lbe) {}
		}
		return list;
	}


	/**
	 * Prints the root nodes to screen for debug purposes
	 *
	 * @param rootNodes the root nodes
	 */
	protected static void printRootNodes(ResolvedConceptReferenceList rootNodes)
	{
		if (rootNodes == null || rootNodes.getResolvedConceptReferenceCount() == 0) return;
		for (Iterator<ResolvedConceptReference> nodes = rootNodes.iterateResolvedConceptReference(); nodes.hasNext(); )
		{
			ResolvedConceptReference ref = nodes.next();
			String code = ref.getConceptCode();
			String name = ref.getEntityDescription().getContent();
			displayMessage("Root: " + code + "(" + name + ")");
		}
	}


	/**
	 * Prints the nodes to screen, for debug purposes.
	 *
	 * @param hmap the hmap
	 */
	protected static void printNodes(HashMap hmap)
	{
		Set keyset = hmap.keySet();
		Iterator it = keyset.iterator();
		while (it.hasNext())
		{
			String asso = (String) it.next();
			ResolvedConceptReferenceList rcrl = (ResolvedConceptReferenceList) hmap.get(asso);
			if (rcrl.getResolvedConceptReferenceCount() == 0) return;
			for (Iterator<ResolvedConceptReference> nodes = rcrl.iterateResolvedConceptReference(); nodes.hasNext(); )
			{
				ResolvedConceptReference ref = nodes.next();
				String code = ref.getConceptCode();
				String name = ref.getEntityDescription().getContent();
				displayMessage("\t" + asso + ": " + code + "(" + name + ")");
			}
		}
	}


	/**
	 * Prints the nodes to PrintWriter
	 *
	 * @param l_pw the pw
	 * @param hmap the hmap
	 */
	protected static void printNodes(PrintWriter l_pw, HashMap hmap)
	{
		Set keyset = hmap.keySet();
		Iterator it = keyset.iterator();
		while (it.hasNext())
		{
			String asso = (String) it.next();
			ResolvedConceptReferenceList rcrl = (ResolvedConceptReferenceList) hmap.get(asso);
			if (rcrl.getResolvedConceptReferenceCount() == 0) return;
			for (Iterator<ResolvedConceptReference> nodes = rcrl.iterateResolvedConceptReference(); nodes.hasNext(); )
			{
				ResolvedConceptReference ref = nodes.next();
				String code = ref.getConceptCode();
				String name = ref.getEntityDescription().getContent();
				displayMessage(l_pw, "\t" + asso + ": " + code + "(" + name + ")");
			}
		}
	}

	/**
	 * Gets the child nodes.
	 *
	 * @param lbSvc the LexBIGService
	 * @param scheme the scheme
	 * @param sab the sab
	 * @param csvt the CodingSchemeVersionOrTag
	 * @param code the code
	 *
	 * @return the child nodes
	 */
	public static HashMap getChildNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code)
	{
		// HashMap key: assoc, value ResolvedConceptReferenceList
		return getAssociatedNodes(
			lbSvc,
			scheme,
			sab,
			csvt,
			code,
			true);
	}

	/**
	 * Gets the parent nodes.
	 *
	 * @param lbSvc the LexBIGService
	 * @param scheme the scheme
	 * @param sab the sab
	 * @param csvt the CodingSchemeVersionOrTag
	 * @param code the code
	 *
	 * @return the parent nodes
	 */
	public static HashMap getParentNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code)
	{
		return getAssociatedNodes(
			lbSvc,
			scheme,
			sab,
			csvt,
			code,
			false);
	}

	protected static HashMap getAssociatedNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code,
			boolean navigateForward)

	{
		ResolvedConceptReferenceList rcrl = null;
		//LexBIGServiceConvenienceMethods lbscm = null;
		HashMap<String,ResolvedConceptReferenceList> hmap = new HashMap<String,ResolvedConceptReferenceList>();
		try {
			//lbscm = (LexBIGServiceConvenienceMethods) lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");
            //GF12808 [KLO, 04/21/08]
			LexBIGServiceConvenienceMethodsImpl lbscm = (LexBIGServiceConvenienceMethodsImpl)
				 lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");
			lbscm.setLexBIGService(lbSvc);


			String[] assocNames = BROADENING_HIER_ASSOC;
			for (int i = 0; i < assocNames.length; i++)
			{
				String assocToCheck = assocNames[i];
				Association subsume = null;
				try {
					subsume = navigateForward
					? getChildrenBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false)
					: getParentsBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false);
				} catch (Exception ex) {

				}

				if (subsume != null) {
					AssociatedConceptList subsumedNodes = subsume.getAssociatedConcepts();
					if (subsumedNodes.getAssociatedConceptCount() > 0)
					{
						if (hmap.get(assocToCheck) != null)
						{
							ResolvedConceptReferenceList list = (ResolvedConceptReferenceList) hmap.get(assocToCheck);
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								list.addResolvedConceptReference(subsumed.next());
							}
						}
						else
						{
							rcrl = new ResolvedConceptReferenceList();
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								rcrl.addResolvedConceptReference(subsumed.next());
							}
							hmap.put(assocToCheck, rcrl);
						}
					}
				}
			}

			assocNames = NARROWING_HIER_ASSOC;
			for (int i = 0; i < assocNames.length; i++)
			{
				String assocToCheck = assocNames[i];
				Association subsume = null;
				try {
				    subsume = navigateForward
					? getParentsBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false)
					: getChildrenBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false);
				} catch (Exception ex) {

				}

				if (subsume != null) {
					AssociatedConceptList subsumedNodes = subsume.getAssociatedConcepts();
					if (subsumedNodes.getAssociatedConceptCount() > 0)
					{
						if (hmap.get(assocToCheck) != null)
						{
							ResolvedConceptReferenceList list = (ResolvedConceptReferenceList) hmap.get(assocToCheck);
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								list.addResolvedConceptReference(subsumed.next());
							}
						}
						else
						{
							rcrl = new ResolvedConceptReferenceList();
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								rcrl.addResolvedConceptReference(subsumed.next());
							}
							hmap.put(assocToCheck, rcrl);
						}
					}
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
		}

		return hmap;
	}


	/**
	 * Return immediate children of the given node within a source SAB
	 * when navigating the specified association.
	 *
	 * @param lbs the LexBIGService
	 * @param lbscm the LexBIGServiceConvenienceMethods
	 * @param sab the sab
	 * @param conceptCode the concept code
	 * @param relationship the relationship
	 * @param association the association
	 * @param codingScheme the coding scheme
	 * @param versionOrTag the CodingSchemeVersionOrTag
	 * @param buildReferencedEntries whether to build referenced entries
	 *
	 *
	 * @return the children by source
	 *
	 * @throws LBException the LB exception
	 */
	protected static Association getChildrenBySource(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{

		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		NameAndValueList sourceQualifiers = null;
		cng = cng.restrictToSourceCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));

		if (association != null && association.trim().length() > 0 && !association.equals("*"))
		{
			if (sab != null && sab.trim().length() > 0)
			{
				sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
			}
			cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);
	    }

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}

        ResolvedConceptReferenceList rcrl = null;
        try {
		    rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), true, false, buildRefEntries, 1, null, null, null, 0);
		} catch (Exception ex) {
            log.warn("Exception thrown - cng.resolveAsList.");
		}

		if (rcrl != null && rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getSourceOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}

	/**
	 * Return immediate parents of the given node within a source SAB
	 * when navigating the given association.
	 *
	 * @param lbs the LexBIGService
	 * @param lbscm the LexBIGServiceConvenienceMethods
	 * @param sab the sab
	 * @param conceptCode the concept code
	 * @param relationship the relationship
	 * @param association the association
	 * @param codingScheme the coding scheme
	 * @param versionOrTag the CodingSchemeVersionOrTag
	 * @param buildReferencedEntries whether to build referenced entries
	 *
	 *
	 * @return the parents by source
	 *
	 * @throws LBException the LB exception
	 */
	protected static Association getParentsBySource(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		NameAndValueList sourceQualifiers = null;//Constructors.createNameAndValueList(sab, "Source");
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);

		cng = cng.restrictToTargetCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));

		if (association != null && association.trim().length() > 0 && !association.equals("*"))
		{
			if (sab != null && sab.trim().length() > 0)
			{
				sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
			}
			cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);
	    }

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}

        ResolvedConceptReferenceList rcrl = null;
        try {
		    rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), false, true, buildRefEntries, 1, null, null, null, 0);
		} catch (Exception ex) {
            log.warn("Exception thrown - cng.resolveAsList.");
		}
		if (rcrl != null && rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getTargetOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}


	static boolean isBlank(String s) {
		if (s == null || s.trim().length() == 0) return true;
		return false;
	}

	static boolean isNotBlank(String s) {
		if (s != null && s.trim().length() > 0) return true;
		return false;
	}

	protected static ResolvedConceptReferenceList getTopNodes(
		    LexBIGService lbs,
			String codingScheme,
			CodingSchemeVersionOrTag versionOrTag,
			String relationContext,
			String association,
			String associatedSAB) throws LBException
	{
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag,
				relationContext);
		if (isNotBlank(association)) {
			ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
			cng = cng.restrictToAssociations(
				lbConvMethods.createNameAndValueList(association),
				isBlank(associatedSAB) ? null
					: lbConvMethods.createNameAndValueList(associatedSAB, "Source"));
		}

		ResolvedConceptReferenceList list = null;
		try {
			list = cng.resolveAsList(null, true, false, 0, 0, null, null, null, 0);
		} catch (Exception e) {
			log.warn("Exception thrown - cng.resolveAsList.");
		}
        return list;
	}

	/**
	 * Retrieves the nodes attached to the system-designated
	 * end node ('@@') for the given association and source SAB.
	 *
	 * @param codingScheme the coding scheme
	 * @param versionOrTag the version or tag
	 * @param relationContext the relation context
	 * @param association the association
	 * @param associatedSAB the associated sab
	 * @param lbs the lbs
	 *
	 * @return the end nodes
	 *
	 * @throws LBException the LB exception
	 */
	protected static ResolvedConceptReferenceList getEndNodes(
		    LexBIGService lbs,
			String codingScheme,
			CodingSchemeVersionOrTag versionOrTag,
			String relationContext,
			String association,
			String associatedSAB) throws LBException
	{
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag,
				relationContext);
		if (isNotBlank(association)) {
			ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
			cng = cng.restrictToAssociations(
				lbConvMethods.createNameAndValueList(association),
				isBlank(associatedSAB) ? null
					: lbConvMethods.createNameAndValueList(associatedSAB, "Source"));
		}
		ResolvedConceptReferenceList list = null;
		try {
		    list = cng.resolveAsList(null, false, true, 0, 0, null, null, null, 0);
		} catch (Exception e) {
			log.warn("Exception thrown - cng.resolveAsList.");
		}
        return list;
	}

	private static void displayMessage(String text)
	{
		log.warn(text);
		pw.println(text);
	}

	private static void displayMessage(PrintWriter l_pw, String text)
	{
		log.warn(text);
		l_pw.println(text);
	}


	/**
	 * Open print writer.
	 *
	 * @param outputfile the outputfile
	 *
	 * @return the prints the writer
	 */
	public PrintWriter openPrintWriter(String outputfile)
	{
		try {
			PrintWriter printw = new PrintWriter (new BufferedWriter(new FileWriter(outputfile)));
            return printw;
		}catch(Exception e) {
			e.printStackTrace();
			return null;
		}
    }

	/**
	 * Close writer.
	 *
	 * @param l_pw the pw
	 */
	public void closeWriter(PrintWriter l_pw)
	{
		if (l_pw == null)
		{
			log.warn("WARNING: closeWriter is not open.");
			return;
		}
		l_pw.close();
	}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////

	protected static Association getAssociationSourceBySAB(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{

		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		NameAndValueList sourceQualifiers = null;
		cng = cng.restrictToSourceCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));

		if (association != null && association.trim().length() > 0 && !association.equals("*"))
		{
			if (sab != null && sab.trim().length() > 0)
			{
				sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
			}
			cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);
	    }

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}

        ResolvedConceptReferenceList rcrl = null;
        try {
			rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), true, false, buildRefEntries, 1, null, null, null, 0);
		} catch (Exception ex) {
			log.warn("Exception thrown - cng.resolveAsList.");
		}
		if (rcrl != null && rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getSourceOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}


	protected static Association getAssociationTargetBySAB(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		NameAndValueList sourceQualifiers = null;//Constructors.createNameAndValueList(sab, "Source");
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);

		cng = cng.restrictToTargetCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));

		if (association != null && association.trim().length() > 0 && !association.equals("*"))
		{
			if (sab != null && sab.trim().length() > 0)
			{
				sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
			}
			cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);
	    }

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}

        ResolvedConceptReferenceList rcrl = null;
        try {
		    rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), false, true, buildRefEntries, 1, null, null, null, 0);
		} catch (Exception ex) {
            log.warn("Exception thrown - cng.resolveAsList.");
		}
		if (rcrl != null && rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getTargetOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	/**
 * Gets the associated nodes.
 *
 * @param lbSvc the LexBIGService
 * @param scheme the scheme
 * @param sab the sab
 * @param csvt the CodingSchemeVersionOrTag
 * @param code the code
 * @param assocNames the assoc names
 * @param navigateForward whether to navigate forward (true) or back
 *
 * @return the associated nodes
 */
public static HashMap getAssociatedNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code,
			String[] assocNames,
			boolean navigateForward)

	{
		ResolvedConceptReferenceList rcrl = null;
		//LexBIGServiceConvenienceMethods lbscm = null;
		HashMap hmap = new HashMap();
		try {
			//GF12808 [KLO, 04/21/08]
			//lbscm = (LexBIGServiceConvenienceMethods) lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");
			LexBIGServiceConvenienceMethodsImpl lbscm = (LexBIGServiceConvenienceMethodsImpl)
				 lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");
			lbscm.setLexBIGService(lbSvc);

			if (assocNames == null)
			{
				try {
					CodingScheme cs = lbSvc.resolveCodingScheme(scheme, csvt);
					if (cs == null) return hmap;
					Vector<String> v = new Vector<String>();
					SupportedAssociation[] assos = cs.getMappings().getSupportedAssociation();
					for (int i=0; i<assos.length; i++)
					{
						 SupportedAssociation sa = (SupportedAssociation) assos[i];
						 v.add(sa.getLocalId());
					}
					assocNames = new String[v.size()];
					for (int i=0; i<v.size(); i++)
					{
						assocNames[i] = (String) v.elementAt(i);
					}

				} catch (Exception e) {
					e.printStackTrace();
					return hmap;
				}
			}

			for (int i = 0; i < assocNames.length; i++)
			{
				String assocToCheck = assocNames[i];

				Association subsume = null;
				try {
					subsume = navigateForward
						? getChildrenBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false)
						: getParentsBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false);
				} catch (Exception ex) {

				}

				if (subsume != null) {
					AssociatedConceptList subsumedNodes = subsume.getAssociatedConcepts();
					if (subsumedNodes.getAssociatedConceptCount() > 0)
					{
						if (hmap.get(assocToCheck) != null)
						{
							ResolvedConceptReferenceList list = (ResolvedConceptReferenceList) hmap.get(assocToCheck);
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								list.addResolvedConceptReference(subsumed.next());
							}
						}
						else
						{
							rcrl = new ResolvedConceptReferenceList();
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								rcrl.addResolvedConceptReference(subsumed.next());
							}
							hmap.put(assocToCheck, rcrl);
						}
					}
				}
			}
		} catch (Exception e) {
			//e.printStackTrace();
			log.warn("Exception thrown - cng.resolveAsList.");
		}
		return hmap;
	}


	/**
	 * Gets the associated nodes.
	 *
	 * @param lbSvc the LexBIGService
	 * @param scheme the scheme
	 * @param sab the sab
	 * @param csvt the CodingSchemeVersionOrTag
	 * @param code the code
	 * @param association the association
	 * @param navigateForward whether to navigate forward (true) or back
	 *
	 * @return the associated nodes
	 */
	public static ResolvedConceptReferenceList getAssociatedNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code,
			String association,
			boolean navigateForward)
	{
		if (association != null && !association.equals("*")) {
			String[] assocNames = new String[] {association};
			HashMap hmap = getAssociatedNodes(
				lbSvc,
				scheme,
				sab,
				csvt,
				code,
				assocNames,
				navigateForward);
			if (hmap != null && hmap.containsKey(association)) return (ResolvedConceptReferenceList) hmap.get(association);
			return new ResolvedConceptReferenceList();
	    }
	    else
	    {
			String[] a = null;
			HashSet hset = new HashSet();
			HashMap hmap = getAssociatedNodes(
				lbSvc,
				scheme,
				sab,
				csvt,
				code,
				a,
				navigateForward);

			ResolvedConceptReferenceList list = new ResolvedConceptReferenceList();
			Iterator it = hmap.keySet().iterator();
			while (it.hasNext())
			{
				String associationNm = (String) it.next();
				ResolvedConceptReferenceList rcrl = (ResolvedConceptReferenceList) hmap.get(associationNm);
				for (Iterator<ResolvedConceptReference> nodes = rcrl.iterateResolvedConceptReference(); nodes.hasNext(); )
				{
					ResolvedConceptReference ref = nodes.next();
					String conceptcode = ref.getConceptCode();
					if (!hset.contains(conceptcode))
					{
						list.addResolvedConceptReference(ref);
					}
				}
			}
			return list;
		}
	}


    //112607
 	/**
     * Gets the navigation direction.
     *
     * @param lbs the lexbig service
     * @param ontologyDisplayLabel the ontology display label
     * @param csvt the coding scheme version or tag
     * @param sab the aource abbreviation
     *
     * @return the navigation direction
     */
    public static boolean getNavigationDirection(LexBIGService lbs, String ontologyDisplayLabel, CodingSchemeVersionOrTag csvt, String sab)
    {
		boolean forward = true;
		try {
//			ConvenienceMethods lbscm = new ConvenienceMethods(lbs);
			ResolvedConceptReferenceList rootNodes = null;
			try {
				rootNodes = getTopNodes(lbs, ontologyDisplayLabel, csvt, null, null, sab);
			} catch (Exception e) {

			}
			if (rootNodes == null || rootNodes.getResolvedConceptReferenceCount() == 0) {
				forward = false;
			}
	    } catch (Exception e) {

		}
		return forward;
	}

}
