package gov.nih.nci.system.client.proxy;

import gov.nih.nci.evs.annotation.LazyLoad;
import gov.nih.nci.system.applicationservice.ApplicationService;
import gov.nih.nci.system.applicationservice.EVSApplicationService;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Vector;

import java.util.HashMap;
import java.util.Iterator;


import net.sf.cglib.proxy.Enhancer;

import org.LexGrid.LexBIG.Impl.logging.LoggerFactory;
import org.LexGrid.annotations.LgAdminFunction;
import org.LexGrid.annotations.LgClientSideSafe;
import org.aopalliance.intercept.MethodInvocation;
import org.apache.log4j.Logger;
import org.springframework.aop.framework.Advised;
import org.springframework.aop.framework.ProxyFactory;


import gov.nih.nci.evs.security.SecurityToken;

/**
 * Object proxy implementation for EVS. Certain methods are overridden to
 * provide EVS-specific proxying functionality.
 *
 * @author <a href="mailto:muhsins@mail.nih.gov">Shaziya Muhsin</a>
 * @author <a href="mailto:rokickik@mail.nih.gov">Konrad Rokicki</a>
 */
public class EVSProxyHelperImpl extends ProxyHelperImpl {

    private static final Logger log = Logger.getLogger(EVSProxyHelperImpl.class);
    static {
        // must configure LexBig before attempting to create any proxies
        LoggerFactory.setLightweight(true);
    }

    /**
     * Annotation class used to mark LexBig classes and methods as
     * safe for execution on a client without the LexBig environment.
     */
    private static final Class CLIENT_SAFE = LgClientSideSafe.class;

    /**
     * Annotation class used to mark LexBig methods which are admin
     * functions and thus illegal for execution via the distributed API.
     */
    private static final Class ADMIN_FUNCTION = LgAdminFunction.class;

    @Override
    protected Object convertCollectionToProxy(ApplicationService as,
            Collection collection) {
        if (null == collection) return null;
        Collection<Object> modifiedCollection = null;
        if (collection instanceof java.util.Vector) {
            modifiedCollection = new Vector<Object>();
        }
        else if (collection instanceof java.util.ArrayList) {
            modifiedCollection = new ArrayList<Object>();
        }
        for (Object obj : collection)
            modifiedCollection.add(convertToProxy(as, obj));
        return modifiedCollection;
    }

    @Override
    protected Object convertObjectToProxy(ApplicationService as, Object obj) {
        if(null == obj) return null;
        if(obj instanceof Integer || obj instanceof Float || obj instanceof Double
                || obj instanceof Character || obj instanceof Long || obj instanceof Boolean
                || obj instanceof String || obj instanceof Date || obj instanceof BeanProxy)
            return obj;

        // Don't proxy client-safe LexBig objects
        if (isLexBigClass(obj.getClass()) && isClientSafe(obj.getClass())) {
            return obj;
        }

        ProxyFactory pf = new ProxyFactory(obj);
        pf.setProxyTargetClass(true);
        pf.addAdvice(new BeanProxy(as, this));
        return pf.getProxy();
    }

    /**
     * Returns true if the object is initialized
     */@SuppressWarnings("unchecked")
    @Override
    public boolean isInitialized(MethodInvocation invocation) throws Throwable {
        Class implClass = invocation.getThis().getClass();

        // LexBig objects have methods that must be executed remotely
        if (isLexBigClass(implClass)) {
            Method method = invocation.getMethod();
            Method methodImpl = implClass.getMethod(method.getName(),
                method.getParameterTypes());

            if (methodImpl.isAnnotationPresent(ADMIN_FUNCTION)) {
                throw new UnsupportedOperationException(
                    "Admin functions cannot be executed using the distributed API");
            }

            if (isClientSafe(methodImpl)) {
                log.info("DLB calling locally: "+implClass.getName()+"."+methodImpl.getName());
                return true;
            }

            log.info("DLB calling remotely: "+implClass.getName()+"."+methodImpl.getName());
            return false;
        }

        Object retVal = invocation.proceed();
        if (invocation.getMethod().isAnnotationPresent(LazyLoad.class)
                && retVal == null) {
            return false;
        }
        return true;
    }
    /**
     * Implements the LazyLoading
     */@SuppressWarnings("unchecked")
    @Override
    public Object lazyLoad(ApplicationService as, MethodInvocation invocation)
            throws Throwable {

        Object bean = invocation.getThis();
        Method method = invocation.getMethod();
        String methodName = method.getName();

        Object args[] = invocation.getArguments();
        Class implClass = bean.getClass();

        // LexBig objects have methods that must be executed remotely
        if (isLexBigClass(implClass)) {

            // TODO: remove impls?
            Method methodImpl = implClass.getMethod(methodName,
                method.getParameterTypes());

            for(int i=0; i<args.length; i++) {
                if (args[i] != null) {
                    if (Enhancer.isEnhanced(args[i].getClass())) {
                        args[i] = unwrap(args[i]);
                    }
                }
            }

            EVSApplicationService eas = (EVSApplicationService)as;
            return eas.executeRemotely(invocation.getThis(),
                methodImpl.getName(),  getParameterTypes(methodImpl), args);
        }

        if (methodName.startsWith("get") && (args == null || args.length == 0)) {

            String vocabularyName = null;
            String fieldName = methodName.substring(3);

            fieldName = Character.toLowerCase(fieldName.charAt(0))
                    + fieldName.substring(1);

            EVSApplicationService eas = (EVSApplicationService) as;
            HashMap securityToken_map = eas.getSecurityToken_map();

            int k = 0;
            SecurityToken token = null;

            Iterator ite = securityToken_map.keySet().iterator();
            while (ite.hasNext())
            {
			   k++;
			   String key = (String) ite.next();
			   Object value = (SecurityToken) securityToken_map.get(key);
			   //System.out.println("\tEVSProxHelperImpl=== key ============ " + key);
			   if (k == 1)
			   {
				   if (value != null)
				   {
					    vocabularyName = (String) key;
				   		token = (SecurityToken) value;
				   }
			   }
		    }
            //Object obj = as.getAssociation(createClone(bean), fieldName);

            EVSApplicationService e = (EVSApplicationService) as;
            Object obj = e.getAssociation(createClone(bean), fieldName, token);
            Object value = obj;

            Field field = getField(bean, fieldName);
            String className = field.getType().getName();
            if (obj instanceof ListProxy) ((ListProxy) obj).setAppService(as);

/*
            if (!(className.equalsIgnoreCase( "java.util.Collection")
                    || className.equalsIgnoreCase("java.util.Vector")
                    ||className.equalsIgnoreCase("java.util.ArrayList"))) {
*/
//KLO, 021108, GF#11889

            if (!(className.equalsIgnoreCase( "java.util.Collection")
                    || className.equalsIgnoreCase("java.lang.Boolean")
                    || className.equalsIgnoreCase("java.util.Vector")
                    ||className.equalsIgnoreCase("java.util.ArrayList"))) {
//                Class<?> x = field.getType();
                Collection results = (Collection) obj;
                if (results.size() == 0)
                    value = null;
                /* [#14273] testGetHistoryRecord throws an exception
                else
                    throw new Exception(
                            "Invalid data obtained from the database for the "
                                    + fieldName + " attribute of the "
                                    + bean.getClass().getName());
                */
            }

            Class[] params = new Class[] { field.getType() };
            Method setter = getMethod(bean, "set"
                    + method.getName().substring(3), params);

            //[#14273] testGetHistoryRecord throws an exception
            if (value != null)
            {
				if (setter != null && params.length == 1) {
					if (method.getReturnType().getName().equals("java.util.Vector")) {
						Vector vec = new Vector();
						for (Object o : (Collection) value) {
							vec.add(o);
						}
						setter.invoke(bean, new Object[] { vec });
						return vec;
					}
					else if (method.getReturnType().getName().equals(
						"java.util.ArrayList")) {
						ArrayList alist = new ArrayList();
						for (Object o : (Collection) value) {
							alist.add(o);
						}
						setter.invoke(bean, new Object[] { alist });
						return alist;
					}

					//KLO, 021108, GF#11889
					else if (method.getReturnType().getName().equals(
						"java.lang.Boolean")) {
						ArrayList alist = (java.util.ArrayList) value;
						java.lang.Boolean aBoolean = (Boolean) alist.get(0);
						setter.invoke(bean, aBoolean);
						return aBoolean;
					}

                    //[#14273] testGetHistoryRecord throws an exception
					else if (method.getReturnType().getName().equals(
						"gov.nih.nci.evs.domain.HistoryRecord")) {
						ArrayList alist = (java.util.ArrayList) value;
						gov.nih.nci.evs.domain.HistoryRecord aHistoryRecord = (gov.nih.nci.evs.domain.HistoryRecord) alist.get(0);
						setter.invoke(bean, aHistoryRecord);
						return aHistoryRecord;
					}

					else {
						setter.invoke(bean, new Object[] { value });
						return value;
					}
				}
		    }
        }

        return null;
    }

    /**
     * Returns true if the given invocation is for a LexBig object.
     *
     * @param clazz the clazz
     *
     * @return true, if checks if is lex big class
     */
    private boolean isLexBigClass(Class clazz) {
        return clazz.getName().startsWith("org.LexGrid");
    }

    /**
     * Returns true if the given method or class is marked client-safe.
     *
     * @param object the object
     *
     * @return true, if checks if is client safe
     */@SuppressWarnings("unchecked")
    private boolean isClientSafe(Object object) {
        if (object instanceof Method) {
            return ((Method)object).isAnnotationPresent(CLIENT_SAFE);
        }
        else {
            return ((Class)object).isAnnotationPresent(CLIENT_SAFE);
        }
    }

    /**
     * Returns the underlying object that the specified proxy is advising.
     *
     * @param proxy the proxy
     *
     * @return the object
     *
     * @throws Exception the exception
     */
    private Object unwrap(Object proxy) throws Exception {
        Object interceptor = null;
        int i = 0;
        while (true) {
            Field field = proxy.getClass().getDeclaredField("CGLIB$CALLBACK_"+i);
            field.setAccessible(true);
            Object value = field.get(proxy);
            if (value.getClass().getName().contains("EqualsInterceptor")) {
                interceptor = value;
                break;
            }
            i++;
        }

        Field field = interceptor.getClass().getDeclaredField("advised");
        field.setAccessible(true);
        Advised advised = (Advised)field.get(interceptor);
        Object realObject = advised.getTargetSource().getTarget();
        return realObject;
    }

    /**
     * Returns a list of class names that are parameters to the given method.
     * @param methodImpl
     * @return list of fully-qualified class names
     */
    private String[] getParameterTypes(Method methodImpl) {
        String[] paramClasses =
            new String[methodImpl.getParameterTypes().length];
        int i = 0;
        for(Class paramClass : methodImpl.getParameterTypes()) {
            if (paramClass == null) continue;
            paramClasses[i++] = paramClass.getName();
        }
        return paramClasses;
    }

    /**
     * Creates a serializable copy of a given object
     */
    protected Object createClone(Object source)
    {
        try
        {
            Object target = source.getClass().newInstance();
            List<Field> fieldList = new ArrayList<Field>();
            getAllFields(source.getClass(), fieldList);
            for(Field field:fieldList)
            {
                Object obj = field.get(source);

                if(obj instanceof Integer || obj instanceof Float || obj instanceof Double
                        || obj instanceof Character || obj instanceof Long || obj instanceof Boolean
                        || obj instanceof String || obj instanceof gov.nih.nci.evs.domain.Vocabulary)
                {
                    if(!Modifier.isStatic(field.getModifiers()) && !Modifier.isFinal(field.getModifiers()))
                    {
                        field.setAccessible(true);
                        field.set(target, obj);
                    }
                }
            }
            return target;
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InstantiationException e) {
            e.printStackTrace();
        }
        return null;
    }
}
